module type S = sig
  type 'a t
  val length : 'a t -> int
  val empty : unit  ->'a t
  val cons : 'a -> 'a t -> 'a t
  val singleton : 'a -> 'a t
  val append : 'a t -> 'a t -> 'a t
  val tabulate : (int -> 'a) -> int -> 'a t
  val nth : 'a t -> int -> 'a
  val filter : ('a -> bool) -> 'a t -> 'a t
  val map : ('a -> 'b) -> 'a t -> 'b t
  val reduce : ('a -> 'a -> 'a) -> 'a -> 'a t -> 'a
  val map_reduce : ('a -> 'b) -> 'b -> ('b -> 'b -> 'b) -> 'a t -> 'b
  val repeat : 'a -> int -> 'a t
  val flatten : 'a t t -> 'a t
  val zip : ('a t * 'b t) -> ('a * 'b) t
  val split : 'a t -> int -> 'a t * 'a t
end

(*******************************************************)
(* Sequential Sequences Based on a List Representation *)
(*******************************************************)

module ListSeq : S = struct

type 'a t = 'a list

let length = List.length

let empty () = []

let cons (x:'a) (s:'a t) = x::s

let singleton x = [x]

let append = List.append

let tabulate f n =
  let rec helper acc x =
    if x = n then List.rev acc
    else helper ((f x)::acc) (x+1) in
  helper [] 0

let nth = List.nth

let filter = List.filter

let map = List.map

let reduce = List.fold_left

let map_reduce l e n s = reduce n e (map l s)

let repeat x n =
  let rec helper x n acc =
    if n = 0 then acc else helper x (n-1) (x::acc) in
  helper x n []

let flatten = List.flatten

let zip (s1,s2) = List.combine s1 s2

let split s i =
  let rec helper s i acc =
    match s,i with
      | _,0 -> (List.rev acc,s)
      | [],_ -> failwith "split"
      | h::t,_ -> helper t (i-1) (h::acc) in
  helper s i []

end

(**********************)
(* Parallel Sequences *)
(**********************)

module ParSeq : S = ListSeq (* Change this *)
